from __future__ import annotations

from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union

import structlog

from . import ff_utils


logger = structlog.getLogger(__name__)

JsonObject = Mapping[str, Any]
LinkTo = Union[str, JsonObject]


def get_link_to(
    existing_item: PortalItem,
    link_to: LinkTo,
    item_to_create: PortalItem,
) -> Union[str, PortalItem]:
    """Create new item model from existing one for given linkTo.

    LinkTos be identifiers (UUIDs) or (partially) embedded objects.

    Follow rules of existing item model for fetching linkTo via
    request. If not fetching via request, then make item model from
    existing properties if possible.
    """
    fetch_links = existing_item.should_fetch_links()
    identifier = get_item_identifier(link_to)
    if fetch_links and identifier:
        return item_to_create.from_identifier_and_existing_item(
            identifier, existing_item
        )
    if isinstance(link_to, Mapping):
        return item_to_create.from_properties_and_existing_item(link_to, existing_item)
    return link_to


def get_item_identifier(item: LinkTo) -> str:
    if isinstance(item, str):
        return item
    if isinstance(item, Mapping):
        return item.get(PortalItem.UUID, "")


@dataclass(frozen=True)
class PortalItem:
    ACCESSION = "accession"
    AT_ID = "@id"
    TYPE = "@type"
    UUID = "uuid"

    properties: JsonObject
    auth: Optional[JsonObject] = field(default=None, hash=False)
    fetch_links: Optional[bool] = field(default=False, hash=False)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._uuid})"

    @property
    def _uuid(self) -> str:
        return self.properties.get(self.UUID, "")

    @property
    def _at_id(self) -> str:
        return self.properties.get(self.AT_ID, "")

    @property
    def _accession(self) -> str:
        return self.properties.get(self.ACCESSION, "")

    @property
    def _types(self) -> List[str]:
        return self.properties.get(self.TYPE, [])

    def should_fetch_links(self) -> bool:
        return self.fetch_links

    def get_auth(self) -> Union[JsonObject, None]:
        return self.auth

    def get_properties(self) -> JsonObject:
        return self.properties

    def get_accession(self) -> str:
        return self._accession

    def get_uuid(self) -> str:
        return self._uuid

    def get_at_id(self) -> str:
        return self._at_id

    def get_types(self) -> List[str]:
        return self._types

    def _get_link_tos(
        self, link_tos: Iterable[LinkTo], item_to_create: PortalItem
    ) -> List[Union[str, PortalItem]]:
        return [self._get_link_to(link_to, item_to_create) for link_to in link_tos]

    def _get_link_to(
        self,
        link_to: LinkTo,
        item_to_create: PortalItem,
    ) -> Union[str, PortalItem]:
        return get_link_to(self, link_to, item_to_create)

    @classmethod
    def from_properties(
        cls,
        properties: JsonObject,
        fetch_links: bool = False,
        auth: JsonObject = None,
        **kwargs: Any,
    ) -> PortalItem:
        return cls(properties, fetch_links=fetch_links, auth=auth)

    @classmethod
    def from_identifier_and_auth(
        cls, identifier: str, auth: JsonObject, fetch_links=False, **kwargs: Any
    ) -> PortalItem:
        properties = cls._get_item_via_auth(identifier, auth)
        return cls.from_properties(properties, auth=auth, fetch_links=fetch_links)

    @classmethod
    def _get_item_via_auth(
        cls, identifier: str, auth: JsonObject, add_on: Optional[str] = "frame=object"
    ) -> JsonObject:
        hashable_auth = cls._make_hashable_auth(auth)
        return cls._get_and_cache_item_via_auth(identifier, hashable_auth, add_on)

    @classmethod
    def _make_hashable_auth(cls, auth: Mapping[str, str]) -> Tuple[Tuple[str, str]]:
        """Assuming nothing nested here."""
        return tuple(auth.items())

    @classmethod
    def _undo_make_hashable_auth(
        cls, hashable_auth: Tuple[Tuple[str, str]]
    ) -> JsonObject:
        return dict(hashable_auth)

    @classmethod
    @lru_cache(maxsize=256)
    def _get_and_cache_item_via_auth(
        cls,
        identifier: str,
        hashable_auth: Tuple[Tuple[str, Any]],
        add_on: Optional[str] = None,
    ) -> JsonObject:
        """Save on requests by caching items."""
        auth = cls._undo_make_hashable_auth(hashable_auth)
        try:
            result = ff_utils.get_metadata(identifier, key=auth, add_on=add_on)
        except Exception as e:
            result = {}
            logger.error(f"Error getting metadata for {identifier}: {e}")
        return result

    @classmethod
    def from_identifier_and_existing_item(
        cls, identifier: str, existing_item: PortalItem, **kwargs: Any
    ) -> PortalItem:
        fetch_links = existing_item.should_fetch_links()
        auth = existing_item.get_auth()
        if auth:
            return cls.from_identifier_and_auth(
                identifier, auth, fetch_links=fetch_links
            )
        raise RuntimeError("Unable to fetch given identifier without auth key")

    @classmethod
    def from_properties_and_existing_item(
        cls, properties: JsonObject, existing_item: PortalItem, **kwargs: Any
    ) -> PortalItem:
        fetch_links = existing_item.should_fetch_links()
        auth = existing_item.get_auth()
        return cls.from_properties(properties, fetch_links=fetch_links, auth=auth)


@dataclass(frozen=True)
class NestedProperty:
    properties: JsonObject
    parent_item: Optional[PortalItem] = field(default=None, hash=False)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(parent={self.parent_item.__repr__()})"

    def get_properties(self) -> JsonObject:
        return self.properties

    def get_parent_item(self) -> Union[PortalItem, None]:
        return self.parent_item

    def _get_link_tos(
        self, link_tos: LinkTo, item_to_create: PortalItem
    ) -> List[Union[str, PortalItem]]:
        return [self._get_link_to(link_to, item_to_create) for link_to in link_tos]

    def _get_link_to(
        self,
        link_to: LinkTo,
        item_to_create: PortalItem,
    ) -> Union[str, PortalItem]:
        if self.parent_item:
            return get_link_to(self.parent_item, link_to, item_to_create)
        if isinstance(link_to, Mapping):
            return item_to_create.from_properties(link_to)
        return link_to

    @classmethod
    def from_properties(
        cls,
        properties: JsonObject,
        parent_item: Optional[PortalItem] = None,
        **kwargs: Any,
    ) -> NestedProperty:
        return cls(properties, parent_item=parent_item)



import redis
import datetime
from typing import Union
# Low level utilities for working with Redis


def create_redis_client(*, url: str, ping=True) -> redis.Redis:
    """ Creates a Redis client connecting from a host:port or a URL
        Because this creation pings Redis should re-use this as much as possible, but
        you can skip the ping if you're confident it will come up
    """
    r = redis.from_url(url)
    if ping:
        r.ping()
    return r


class RedisException(Exception):
    pass


class RedisBase(object):
    """ This class contains low level methods meant to implement useful Redis APIs. The idea is these functions
        are used to implement the methods needed in Redis Tools.

        One important note about Redis - the database wants bytes like objects in the
        place of strings, so some automatic encoding/decoding from utf-8 is implemented
        here and thus should NOT be done by the caller. If the caller does do it though
        it is assumed it has been done correctly, or is of a type (ie: float) that does
        not require encoding/decoding.

        TODO: implement more APIs as needed - Redis is very complex and offers a lot of
              potentially useful functionality, and it may be necessary to split off into
              sub classes later on to isolate in a reasonable manner, but for now all here
    """

    def __init__(self, redis_handle):
        self.redis = redis_handle

    @staticmethod
    def _encode_value(value: Union[str, int, float]) -> Union[bytes, int, float]:
        """ Function used to preprocess all values that are strings and encode them as utf-8 bytes
        :param value: any value that can be represented as bytes either directly (numbers, floats)
                      or through encoding (string)
        :return: encoded value
        """
        return value.encode('utf-8') if (value and isinstance(value, str)) else value

    @staticmethod
    def _decode_value(value: Union[bytes, int, float]) -> str:
        """ Function used to post process all values that are bytes and decode them into utf-8 strings
        :param value: any of bytes or other single valued entities (floats, integers)
        :return: decoded value
        """
        return value.decode('utf-8') if (value and isinstance(value, bytes)) else value

    def info(self) -> dict:
        """ Returns info about the Redis server https://redis.io/commands/info/
        :return: a dictionary of information about the redis server
        """
        return self.redis.info()

    def set(self, key: str, value: Union[str, int, float], exp: Union[int, datetime.timedelta] = None) -> str:
        """ Sets the given key to the given value https://redis.io/commands/set/
        :param key: string to store value under
        :param value: value mapped to from key
        :param exp: expiration time in seconds as int or datetime.timedelta (optional)
        :return: a "true" string <OK>
        """
        kwargs = {}
        if exp:
            kwargs['ex'] = exp
        return self.redis.set(self._encode_value(key), self._encode_value(value), **kwargs)

    def get(self, key: str) -> str:
        """ Gets the given key from Redis https://redis.io/commands/get/
        :param key: key to check for a value store in Redis
        :return: that value if mapped or None if key does not exist
        """
        val = self.redis.get(self._encode_value(key))
        if val is not None:
            val = self._decode_value(val)
        return val

    def set_expiration(self, key: str, t: Union[int, datetime.time]) -> bool:
        """ Sets the TTL of the given key manually
        :param key: key to set TTL
        :param t: time in seconds until expiration (datetime.timedelta or int)
        :returns: True if successful
        """
        return self.redis.expire(key, t, gt=True)

    def ttl(self, key: str) -> datetime.time:
        """ Gets the TTL of the given key
        :param key: key to get TTL for
        :return: datetime value in seconds
        """
        return self.redis.ttl(key)

    def delete(self, key: str) -> int:
        """ Deletes the given key from Redis https://redis.io/commands/del/
        :param key: key to delete
        :return: number of keys removed
        """
        return self.redis.delete(self._encode_value(key))

    def hget(self, key: str, field: str) -> str:
        """ Gets the value of field from hash key https://redis.io/commands/hget/
        :param key: hash key to retrieve field value from
        :param field: sub key within the hash key whose value we'd like to retrieve
        :return: the value key -> field -> value mapped, if it exists
        """
        return self._decode_value(self.redis.hget(self._encode_value(key), self._encode_value(field)))

    def hgetall(self, key: str) -> dict:
        """ Gets all values of the given hash https://redis.io/commands/hgetall/
        :param key: hash key to grab all values from
        :return: dictionary of values mapped within the given hash key
        """
        encoded_vals = self.redis.hgetall(self._encode_value(key))
        if encoded_vals:
            encoded_vals = {self._decode_value(k): self._decode_value(v) for k, v in encoded_vals.items()}
        return encoded_vals

    def hset(self, key: str, field: str, value: Union[str, int, float]) -> int:
        """ Sets a single field on a hash key https://redis.io/commands/hset/
        :param key: hash key to set field -> value mapping on
        :param field: field to map under the hash key
        :param value: value to map from field
        :return: number of fields set (1 if successful)
        """
        return self.redis.hset(self._encode_value(key), self._encode_value(field), self._encode_value(value))

    def hset_multiple(self, key: str, items: dict) -> int:
        """ Sets all k,v pairs in items on hash key https://redis.io/commands/hset/ (variadic form)
        :param key: hash key to store items under
        :param items: k, v pairs to store under the hash key
        :return: number of k, v pairs set (== len(items.items()) if successful)
        """
        encoded_dict = {self._encode_value(k): self._encode_value(v) for k, v in items.items()}
        return self.redis.hset(key, mapping=encoded_dict)

    def dbsize(self) -> int:
        """ Returns number of keys in redis https://redis.io/commands/dbsize/ """
        return self.redis.dbsize()



from contextlib import contextmanager
from types import ModuleType
from typing import Any, Iterator, Optional
from unittest import mock


AN_UNLIKELY_RETURN_VALUE = "unlikely return value"


@contextmanager
def patch_context(
    to_patch: object,
    return_value: Any = AN_UNLIKELY_RETURN_VALUE,
    module: Optional[ModuleType] = None,
    **kwargs,
) -> Iterator[mock.MagicMock]:
    if isinstance(to_patch, property):
        to_patch = to_patch.fget
        new_callable = mock.PropertyMock
    else:
        new_callable = mock.MagicMock
    if module is None:
        target = f"{to_patch.__module__}.{to_patch.__qualname__}"
    else:
        target = f"{module.__name__}.{to_patch.__qualname__}"
    with mock.patch(target, new_callable=new_callable, **kwargs) as mocked_item:
        if return_value != AN_UNLIKELY_RETURN_VALUE:
            mocked_item.return_value = return_value
        yield mocked_item
#
#
#
# @contextmanager
# def patch_context(
#     object_to_patch: object,
#     attribute_to_patch: str,
#     return_value: Optional[Any] = None,
#     **kwargs,
# ) -> Iterator[mock.MagicMock]:
#     with mock.patch.object(object_to_patch, attribute_to_patch, **kwargs) as mocked_item:
#         if return_value is not None:
#             mocked_item.return_value = return_value
#         yield mocked_item
